/******************************************************************************
 * This file is part of a software distribution, which is furnished under the *
 * terms of a license.  Use of this software  by any means is subject to this *
 * license  and  signifies  the  acceptance of  the  licensing  terms  stated *
 * therein. Please see  the file LICENSE in the  top-level directory  of this *
 * software  distribution  for detailed copyright  disclaimers  and licensing *
 * terms.                                                                     *
 ******************************************************************************
 * Copryight (c) by Andreas S. Wetzel - All rights reserved.                  *
 ******************************************************************************/

/* $Id: vc_input.c,v 1.4 2001/03/19 23:49:08 mickey Exp $ */

#include <vchat.h>
#include <proto_vchat.h>

#include <ctype.h>

/********* Globals *********/

ED	main_ed;

#if SHELLMODE
ED	cmd_ed;
#endif

ED	*ed = &main_ed;

extern int in_chat_window;

extern VP vp;
extern VTCAP vtcap;

/*
 * int input_line(char *prompt, size_t maxlen, u_short flags);
 *
 * Read input from the users terminal up to maxlen characters (or the
 * maximum possible length if maxlen is given as 0) and return the number
 * of bytes actually read into the globel ED buffer pointed to by the global
 * ed variable. Flags should be the bitwise OR of any of the following flags:
 *
 * ED_RST_CTX	- Reset input context upon startup
 * ED_NO_ECHO	- Suppress terminal echo while reading input
 *
 */

int input_line(char *prompt, size_t maxlen, u_short flags)
{
	u_char	c;
#if SHELLMODE
	static ED	*saved_ed	= NULL;
#endif

	/*** Lock all interrupts during init ***/

#if SHELLMODE
	if(saved_ed == NULL)
#endif
		vlock(LOCK, (V_IO | V_TIMER | V_WINCH));

	/*** Disable videosnake ***/

	snake_disable();

	/**** Buffer init ****/

	if(flags & ED_RST_CTX)
	{
		/*
		 * Reset input context upon startup
		 */

		BZERO(&ed->buffer, EDBUFSIZE+1);
		ed->endpos = ed->curpos = ed->margin = (char *) &ed->buffer;
		ed->prompt = NULL;
	}

	if(prompt)
		ed->prompt = prompt;

	ed->offset = (ed->prompt != NULL) ? real_len(ed->prompt) : 0;
	ed->maxlen = (maxlen > 0 && maxlen < EDBUFSIZE) ? maxlen : EDBUFSIZE;
	ed->flags = flags;

	/***** window init *****/

	line_update();

	/*** init completed - unlock interrupts ***/

	vlock(UNLOCK, (V_IO | V_TIMER | V_WINCH));

	/*** Wait for user input ***/

	while((c = getchar()) != CR && c != LF)
	{
		vlock(LOCK, (V_IO | V_TIMER | V_WINCH));

		snake_enable();

		if(is_key_sequence(vtcap.key_cursleft, c))
		{
			curs_left();
		}
		else if(is_key_sequence(vtcap.key_cursright, c))
		{
			curs_right();
		}
		else if(is_key_sequence(vtcap.key_cursup, c))
		{
			if((ed->flags & ED_HISTORY) != 0
			&& vp.history_current != vp.history_start)
			{
				snake_disable();
				vlock(UNLOCK, (V_IO | V_TIMER | V_WINCH));
				return(-2);
			}
			else
			{
				nop();
			}
		}
		else if(is_key_sequence(vtcap.key_cursdown, c))
		{
			if((ed->flags & ED_HISTORY) != 0
			&& vp.history_current != vp.history_next)
			{
				snake_disable();
				vlock(UNLOCK, (V_IO | V_TIMER | V_WINCH));
				return(-4);
			}
			else
			{
				nop();
			}
		}
		else if(is_key_sequence(vtcap.key_insert, c))
		{
			ungetc(' ', stdin);
		}
		else if(is_key_sequence(vtcap.key_delete, c))
		{
			delete();
		}
		else if(iscntrl(c))
		{
			switch(c)
			{
#if SHELLMODE
				case CTL_P:	if(saved_ed == NULL)
						{
							/*
							 * Switch to shell mode
							 */

							u_long	x;
							u_char	prpt[8];

							saved_ed = ed;
							ed = &cmd_ed;

							vp.cmdmode = 1;
							update_flags();


							ed->flags &= ~(ED_SHL_DONE);

							sprintf(prpt, "%s$%s ", S_FG_CYN, S_OFF);

							for(;;)
							{
								x = input_line(prpt, 0, (vp.cmdcontext_invalid) ? ED_RST_CTX : 0);

								if(ed->flags & ED_SHL_DONE)
								{
									vp.cmdcontext_invalid = 0;
									break;
								}

								if(x > 0)
								{
									snd_serv(CMD_CTXTCMD, ed->buffer, strlen(ed->buffer));
									vp.cmdcontext_invalid = 1;
								}
								else
								{
									vp.cmdcontext_invalid = 0;
								}
							}

							ed->flags &= ~(ED_SHL_DONE);
							vp.cmdmode = 0;
							ed = saved_ed;
							saved_ed = (ED *)NULL;

							line_update();
							update_flags();
						}
						else
						{
							/*
							 * Switch to chat mode
							 */

							ed->flags |= ED_SHL_DONE;
							return(0);
						}

						break;
#endif

				case DEL:
				case BS:	backspace();
						break;

				case CTL_A:	jump_start();
						break;

				case CTL_E:	jump_end();
						break;

				case CTL_D:	delete();
						break;

				case CTL_K:	erase_eol();
						break;

				case CTL_U:	erase_line();
						break;

				case CTL_F:	word_forward();
						break;

				case CTL_B:	word_reverse();
						break;

				case CTL_L:	line_update();
						break;

				case HT:	tabcomplete();
						break;

				default:	break;
			}
		}
		else
		{
			if(ed->curpos != ed->endpos)
			{
				/*
				 * Insert mode
				 */

				insert_char(c);
			}
			else
			{
				/*
				 * Append mode
				 */

				append_char(c);
			}
		}

		vlock(UNLOCK, (V_IO | V_TIMER | V_WINCH));
	}

	if(ED_USED)
	{
		vlock(LOCK, (V_IO | V_TIMER | V_WINCH));

		mv(1, vtcap.rows);

		in_chat_window = 0;

		if(vtcap.clear2eol)
			tputs(vtcap.clear2eol, 1, (void *)outc);
		else if(vtcap.clear2eod)
			tputs(vtcap.clear2eod, 1, (void *)outc);

		vlock(UNLOCK, (V_IO | V_TIMER | V_WINCH));
	}

	return(ED_USED);
}
